--Copyright 2016 Shadowsdabom
--This program is free software: you can redistribute it and/or modify
--it under the terms of the GNU General Public License as published by
--the Free Software Foundation, either version 3 of the License, or
--(at your option) any later version.
--
--This program is distributed in the hope that it will be useful,
--but WITHOUT ANY WARRANTY; without even the implied warranty of
--MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--GNU General Public License for more details.
--
--You should have received a copy of the GNU General Public License
--along with this program.  If not, see <http://www.gnu.org/licenses/>.

-------------
--TODO List--
-------------
--TODO: need to check that merchant items listed don't have reputation requirements
--TODO: need to check items for uniqueness limit (i.e. unique(10) means you can only have 10 at a time)
--TODO: need to revert some code to no longer use item.num since wow api buy function buys items seperately regardless of numPerPurchase

-------------------
--State Variables--
-------------------
local isAtMerchant = false; --true when at a merchant, false when not
local merchantUpdated = false;
local merchantItems = {}; --populated with item data when at a merchant

------------------------
--Error Code Constants--
------------------------
local ERROR_INVALID_NUM_ARGS = 1;
local ERROR_INVALID_QUANTITY = 2;
local ERROR_NOT_AT_MERCHANT  = 3;
local ERROR_ITEM_NOT_FOUND   = 4;
local ERROR_MISSING_CURRENCY = 5;
local ERROR_NOT_ENOUGH_ROOM  = 6;
local ERROR_ORDER_FAILED     = 7;

-------------
--Constants--
-------------
local MAX_NUM_BAGS = 4; --wow doesn't count backback as a bag

----------------------------
--Shared Library Functions--
----------------------------
--returns true if the first string starts with the second string, false otherwise
local function startsWith(str, start)
	if (start == "") then
		return true;
	elseif (string.len(start) > string.len(str)) then
		return false;
	end

	return string.sub(str, 1, string.len(start)) == start;
end

--returns string with beginning space and ending space removed
local function trimString(str)
	return str:gsub("^%s*(.-)%s*$", "%1");
end

--returns table populated with substrings found between sep substrings
--taken from post found @http://stackoverflow.com/questions/1426954/split-string-in-lua
local function splitString(inputstr, sep)
        if (sep == nil) then
            sep = "%s";
        end
		
        local t = {};
		
        for str in string.gmatch(inputstr, "([^"..sep.."]+)") do
                table.insert(t, str);
        end
		
        return t;
end

--returns true if the string str is a number, false if not
local function isNumber(str)
	return tonumber(str) ~= nil;
end

--returns true if the string str is an integer, false if not
local function isInteger(str)
	local num = tonumber(str);
	
	if (num == nil) then
		return false;
	else
		return num == math.floor(num);
	end
end

-----------------------------------
--Addon Specific Helper Functions--
-----------------------------------
--prints a message to the ingame console with this addon's color and prefix
local function msg(str)
	print("|cfff9d60eItemBuyer: " .. str);
end

--prints an error message to the ingame console with this addon's color and prefix
local function sendError(num, arg)
	local myError = "|cffff0000[Error " .. num .. "] ";
	
	if (num == ERROR_INVALID_NUM_ARGS) then
		myError = myError .. "Incorrect number of arguments found.";
	elseif (num == ERROR_INVALID_QUANTITY) then
		myError = myError .. "'" .. arg .. "' is not a valid quantity. Must be either a positive integer or valid keyword.";
	elseif (num == ERROR_NOT_AT_MERCHANT) then
		myError = myError .. "Must be at a merchant.";
	elseif (num == ERROR_ITEM_NOT_FOUND) then
		myError = myError .. "The Merchant does not sell '" .. arg .. "'.";
	elseif (num == ERROR_MISSING_CURRENCY) then
		myError = myError .. "Do not have the required currency to complete purchase.";
	elseif (num == ERROR_NOT_ENOUGH_ROOM) then
		myError = myError .. "Do not have the required free space to complete purchase.";
	elseif (num == ERROR_ORDER_FAILED) then
		myError = myError .. "Could not complete purchase; failed to receive order."
	else
		myError = myError .. "An unknown error has occured.";
	end
	
	--sending error to console
	msg(myError);
end

--returns a table of the player's known currencies indexed by name and containing their amount
local function getPlayerCurrency()
	local currency = {};
	
	local numCurrencies = GetCurrencyListSize();
	
	--adding all of the player's known currency
	for i = 1, numCurrencies do
		local name, isHeader, _isExpanded, _isUnused, _isWatched, count, _extraCurrencyType, _icon, itemID = GetCurrencyListInfo(i);
		if (not isHeader) then
			currency[name] = {
				amount = count
			};
		end
	end
	
	--adding copper as a currency
	currency["Money"] = {
		amount = GetMoney()
	};
	
	return currency;
end

--returns a table of the current merchant's item selection indexed by name and containing their index/link/costs
local function getMerchantItems()
	local numItems = GetMerchantNumItems();
	local items = {};
	
	for i = 1, numItems do
		local itemCosts = {};
		local itemName, _texture, itemCopperPrice, numPerPurchase, _numAvailable, _isUsable, hasExtendedCost = GetMerchantItemInfo(i);
		local itemLink = GetMerchantItemLink(i);
		
		numPerPurchase = 1; -- TODO: temporary fix for stack misunderstanding
		
		--if the merchant isn't up to date return an empty list for now, let update handler call us again
		if (itemLink == nil) then
			return {};
		end
		
		local maxPurchaseLimit = GetMerchantItemMaxStack(i);
		local maxStackSize = select(8, GetItemInfo(itemLink));
		
		if (maxStackSize == nil) then
			maxStackSize = 1;
		end
		
		if (hasExtendedCost) then
			local currencyCount = GetMerchantItemCostInfo(i);
			
			for j = 1, currencyCount do
				local _texture, currencyAmount, currencyLink, currencyName = GetMerchantItemCostItem(i, j);
				
				--name is nil for currency as inventory item, currencyLink is nil for normal currency
				if (currencyName == nil and currencyLink ~= nil) then
					currencyName = select(1, GetItemInfo(currencyLink));
				end
				
				local currency = {
					name = currencyName,
					link = currencyLink,
					amount = currencyAmount
				};
				table.insert(itemCosts, currency);
			end
		end
		
		if (itemCopperPrice > 0) then
			local currency = {
				name = "Money",
				link = "",
				amount = itemCopperPrice
			};
			table.insert(itemCosts, currency);
		end
		
		items[itemName] = {
			index = i,
			costs = itemCosts,
			num = numPerPurchase,
			link = itemLink,
			buyLimit = maxPurchaseLimit,
			maxStack = maxStackSize
		};
	end
	
	merchantUpdated = true;
	
	return items;
end

--returns a table of the players inventory including bag info and a list of items (item lists) by item name
local function getPlayerInventory()
	local inventory = {
		numSlots = 0,
		freeSlots = 0,
		bags = {},
		items = {}
	};
	
	local totalFreeSlots = 0;
	local totalNumSlots = 0;
	
	for i = 0, MAX_NUM_BAGS do
		local bagName = GetBagName(i);
		
		if (bagName ~= nil and bagName ~= "") then
			local freeSlots, bagType = GetContainerNumFreeSlots(i);
			local numBagSlots = GetContainerNumSlots(i);
			
			totalFreeSlots = totalFreeSlots + freeSlots;
			totalNumSlots = totalNumSlots + numBagSlots;
			
			for j = 1, numBagSlots do
				local _texture, currentStackSize, _locked, _quality, _readable, _lootable, itemLink, _isFiltered, _hasNoValue, itemID = GetContainerItemInfo(i, j);
				
				--chacking is slot has item in it
				if (itemID ~= nil) then
					local itemName, _link, _quality, _iLevel, _reqLevel, _class, _subclass, maxStackSize, _equipSlot, _texture, _vendorPrice = GetItemInfo(itemID);
					
					if (maxStackSize == nil) then
						maxStackSize = 1;
					end
					
					local numTilFull = maxStackSize - currentStackSize;
					
					--checking if item already exists in set
					if (inventory.items[itemName] == nil) then
						inventory.items[itemName] = {};
					end
					
					table.insert(inventory.items[itemName], {
						bagIndex = i,
						slotIndex = j,
						num = currentStackSize,
						maxNum = maxStackSize,
						link = itemLink,
						id = itemID,
						numLeft = numTilFull
					});
				end
			end
			
			table.insert(inventory.bags, {
				name = bagName,
				numSlots = numBagSlots,
				numFreeSlots = freeSlots,
				family = bagType,
				index = i
			});
		end
	end
	
	return inventory;
end

--helps with debugging by printing a player item's values
local function playerItemPrinter(item)
	print("bagIndex: "  .. tostring(item.bagIndex));
	print("slotIndex: " .. tostring(item.slotIndex));
	print("num: "       .. tostring(item.num));
	print("maxNum: "    .. tostring(item.maxNum));
	print("link: "      .. tostring(item.link));
	print("id: "        .. tostring(item.id));
	print("numLeft: "   .. tostring(item.numLeft));
end

--parses out argument strings into table (1st arg is args[1], 2nd arg is arg[2], etc.)
local function parseArgs(str)
	str = trimString(str);
	local args = {};
	local strLength = string.len(str);
	
	local arg = "";
	local isItemName = false;
	local  itemName = "";
	
	for ch in str:gmatch(".") do
		if (ch == '[') then
			--start of item name
			isItemName = true;
			arg = arg .. ch;
		elseif (isItemName and ch ~= ']') then
			--part of item's name
			arg = arg .. ch;
			itemName = itemName .. ch;
		elseif (isItemName and ch == ']') then
			--end of item name
			isItemName = false;
			arg = arg .. ch;
		elseif (not isItemName and ch ~= ' ') then
			--part of arg
			arg = arg .. ch;
		elseif (not isItemName and ch == ' ') then
		
			if (arg ~= "") then
				if (itemName ~= "") then
					--add item name to arg list instead of item link
					table.insert(args, itemName);
					
					--emptying item name for next possible item
					itemName = "";
				else
					--add arg to args list
					table.insert(args, arg);
				end
				
				--new arg
				arg = "";
			end
		end
	end
	
	--checking if arg was being built before end of string reached
	if (arg ~= "") then
		if (itemName ~= "") then
			--add item name to arg list instead of item link
			table.insert(args, itemName);
		else
			--add arg to args list
			table.insert(args, arg);
		end
	end
	
	return args;
end

--returns a table with the necessary information to complete the requested purchase
--assumes there is enough bag space, currency, and all stacks are a factor of item.num (with exception of 1)
local function getPurchaseOrder(itemName, quantity, playerInventory)
	local item = merchantItems[itemName];
	local bagItems = playerInventory.items[itemName];
	quantity = math.ceil(quantity/item.num) * item.num; --rounding up to next purchasable increment (i.e. buy 6 with groups of 5 means buy 10 to get the 6)
	
	local order = {
		item = item, --referencing original item so we can use item.name, item.link, and item.index
		quantity = quantity,
		purchases = {},
		totalPurchases = 0
	};
	
	--getting purchasing order of items to fill incomplete stacks in bags first
	if (bagItems ~= nil) then
		for i, bagItem in ipairs(bagItems) do
			local numPurchases = math.floor(bagItem.numLeft/item.num); -- numItemLeftForStack/numItemPerPurchase
			local numLeft = numPurchases * item.num;
			
			if (numPurchases > 0 and numLeft <= quantity) then
				table.insert(order.purchases, numPurchases);
				order.totalPurchases = order.totalPurchases + numPurchases;
				quantity = quantity - numLeft;
			end
		end
	end
	
	local numPurchases = math.floor(quantity/item.num);
	
	--converting remaining quantity into remaining purchasing order based on maxBuyLimit per purchase
	while (numPurchases > 0 and item.buyLimit <= numPurchases) do
		table.insert(order.purchases, item.buyLimit);
		order.totalPurchases = order.totalPurchases + item.buyLimit;
		numPurchases = numPurchases - item.buyLimit;
	end
	
	if (numPurchases > 0) then
		table.insert(order.purchases, numPurchases);
		order.totalPurchases = order.totalPurchases + numPurchases;
	end
	
	return order;
end

--fulfills the specified order by making the desired purchases
local function buyItem(order)
	if (order == nil or order.item == nil or order.quantity == nil or order.purchases == nil or order.totalPurchases == nil) then
		sendError(ERROR_ORDER_FAILED);
		return;
	elseif (order.quantity <= 0 or order.totalPurchases <= 0) then
		--no need to do anything else if purchase was for 0 items
		return;
	end
	
	local itemStr = "";
	local itemLink = tostring(order.item.link);
	local numPerPurchase = order.item.num;
	
	if (numPerPurchase ~= nil and numPerPurchase > 1) then
		itemStr = itemLink .. "x" .. tostring(numPerPurchase);
	else
		itemStr = itemLink;
	end
	
	for i, purchaseAmount in ipairs(order.purchases) do
		msg("Buying " .. itemStr .. " x" .. tostring(purchaseAmount));
		BuyMerchantItem(order.item.index, purchaseAmount);
	end
end

--returns true if the player has the correct ammount of currency to purchase the desired quantity of item, false if not
local function hasEnoughCurrency(itemName, quantity, playerInventory)
	local playerCurrency = getPlayerCurrency();
	local item = merchantItems[itemName];
	local numPurchases = math.ceil(quantity/item.num); -- quantity/numItemsPerPurchase
	
	for i, cost in ipairs(item.costs) do
		local neededAmount = cost.amount * numPurchases;
		local currency = playerCurrency[cost.name];
		local currentAmount;
		
		--checking if player even has this currency
		if (currency == nil) then
			currentAmount = 0;
			currency = playerInventory.items[cost.name]; --checking if currency is an inventory item the player has
			
			--player does have this currency in their bags
			if (currency ~= nil) then
				for i, playerItem in ipairs(currency) do
					currentAmount = currentAmount + playerItem.num;
				end
			end
		else
			currentAmount = currency.amount;
		end
		
		local leftover = currentAmount - neededAmount;
		
		if (leftover < 0) then
			return false;
		end
	end
	
	return true;
end

--used to help get the minimum value in the table of max values
local function compareNums(a, b)
	return a < b;
end


--modified version of code @http://wowprogramming.com/docs/api/GetItemFamily
local function itemCanGoInBag(itemLink, bagFamily)
   local itemFamily = GetItemFamily(itemLink);
   local equipSlot = select(9, GetItemInfo(itemLink));
   
   -- If the item is a container, then the itemFamily should be 0
   if (equipSlot == "INVTYPE_BAG") then
      itemFamily = 0;
   end
   
	if (bagFamily == 0) then
		return true;
	elseif (bit.band(itemFamily, bagFamily) > 0) then
		return true;
	end
	
	return false;
end

--returns the max number of items that can fit in the current player's bags
local function getMaxQuantityBagSpace(itemName, playerInventory)
	local numFreeSlots = 0;
	local item = merchantItems[itemName];
	
	--getting the number of open slots in all players bags that this item can fill
	for i, bag in ipairs(playerInventory.bags) do
		if (itemCanGoInBag(item.link, bag.family)) then
			numFreeSlots = numFreeSlots + bag.numFreeSlots;
		end
	end
	
	local numItems = numFreeSlots * item.maxStack;
	local playerItems = playerInventory.items[itemName];
	
	--including incomplete stacks of this item already in players bag
	if (playerItems ~= nil) then
		for i, playerItem in ipairs(playerItems) do
			numItems = numItems	+ playerItem.numLeft;
		end
	end
	
	--making sure numItems is a reachable amount based on numItemPerPurchase factor (i.e. can't buy 7 items if numItemPerPurchase is 5)
	numItems = math.floor(numItems/item.num) * item.num;
	
	return numItems; -- returns number of item that can possibly fit in bags
end

--returns the max number of items that can be bought with the current player's immediate currency
local function getMaxQuantityCurrency(itemName, playerInventory)
	local playerCurrency = getPlayerCurrency();
	local item = merchantItems[itemName];
	local maxQuantities = {};
	
	for i, cost in ipairs(item.costs) do
		local currency = playerCurrency[cost.name];
		local currentAmount;
		
		--checking if player even has this currency
		if (currency == nil) then
			currentAmount = 0;
			currency = playerInventory.items[cost.name]; --checking if currency is an inventory item the player has
			
			--player does have this currency in their bags
			if (currency ~= nil) then
				for i, playerItem in ipairs(currency) do
					currentAmount = currentAmount + playerItem.num;
				end
			end
		else
			currentAmount = currency.amount;
		end
		
		local maxNum = math.floor(currentAmount / cost.amount);
		table.insert(maxQuantities, maxNum);
	end
	
	table.sort(maxQuantities, compareNums);
	
	return maxQuantities[1] * item.num; -- returning max purchases * numItemsPerPurchase
end

--returns the maximum quantity of the specified item that can be bought based on bag space, currency, and item limit
local function getMaxViableQuantity(itemName, playerInventory)
	return math.min(getMaxQuantityCurrency(itemName, playerInventory), getMaxQuantityBagSpace(itemName, playerInventory));
end

--returns true if the player has enough bag space to fit the desired amount of the specific item, false if not
local function canFitItemInBag(itemName, quantity, playerInventory)
	local maxQuantity = getMaxQuantityBagSpace(itemName, playerInventory);
	
	if (quantity <= maxQuantity) then
		return true;
	else
		return false;
	end
end

--Opens a confirmation window to make sure the purchase order is correct
local function ShowConfirmation(order)
	local totalPurchases = order.totalPurchases;
	local itemLink = tostring(order.item.link);
	local str = "";
	
	local numPerPurchase = order.item.num;
	if (numPerPurchase ~= nil and numPerPurchase > 1) then
		str = itemLink .. "x" .. tostring(numPerPurchase);
	else
		str = itemLink;
	end
	
	--Showing confirmation and passing order as argument
	local dialog = StaticPopup_Show("ItemBuyerConfirmation", totalPurchases, str);
	if (dialog) then
		dialog.data = order;
		dialog.data2 = nil;
	end
end

------------------------
--Main Event Listeners--
------------------------
--Listens for the '/itembuyer [itemName|itemLink] [Quantity]' command and executes it
local function itemBuyListener(command, editbox)
	local args = parseArgs(command);
	
	--checking command line errors
	if (#args ~= 2) then
		sendError(ERROR_INVALID_NUM_ARGS);
		--msg("/itembuyer ['['ITEM_NAME']'|ITEM_LINK] [QUANTITY|'ALL']");
		return;
	elseif (strlower(args[2]) ~= "all") then
		if (not isInteger(args[2])) then
			sendError(ERROR_INVALID_QUANTITY, args[2]);
			--msg("/itembuyer ['['ITEM_NAME']'|ITEM_LINK] [QUANTITY|'ALL']");
			return;
		else
			local quantity = tonumber(args[2]);
			
			if (quantity < 0) then
				sendError(ERROR_INVALID_QUANTITY, args[2]);
				--msg("/itembuyer ['['ITEM_NAME']'|ITEM_LINK] [QUANTITY|'ALL']");
				return;
			end
		end
	end
	
	local itemName = args[1];
	local quantity;
	
	local playerInventory = getPlayerInventory();
	
	--Gettign correct quantity incase of keyword used
	if (strlower(args[2]) == "all") then
		quantity = getMaxViableQuantity(itemName, playerInventory);
	else
		quantity = tonumber(args[2]);
	end
	
	--checking resource errors
	if (not isAtMerchant) then
		sendError(ERROR_NOT_AT_MERCHANT);
		return;
	elseif (merchantItems[itemName] == nil) then
		sendError(ERROR_ITEM_NOT_FOUND, itemName);
		return;
	elseif (not hasEnoughCurrency(itemName, quantity, playerInventory)) then
		sendError(ERROR_MISSING_CURRENCY);
		return;
	elseif (not canFitItemInBag(itemName, quantity, playerInventory)) then
		sendError(ERROR_NOT_ENOUGH_ROOM);
		return;
	end
	
	local order = getPurchaseOrder(itemName, quantity, playerInventory);
	
	msg("Item: " .. tostring(order.item.link) .. " Quantity: " .. tostring(order.quantity));
	
	--successfully met all the requirements for fulfillling purchase; asking user if purchase order is correct then buying items
	ShowConfirmation(order);
end

--Listens for the '/itemseller [itemName|itemLink] [Quantity]' command and executes it
local function itemSellListener(command, editbox)
	-- TODO: search inventory for specific item and sell quantity num items of it
	msg("'/itemseller' command might be incorperated in later version");
end

---------------
--Addon Setup--
---------------
--Setting up frame listener for loading of addon
local f = CreateFrame("Frame");
f:RegisterEvent("ADDON_LOADED", "ItemBuyer");
f:RegisterEvent("MERCHANT_SHOW");
f:RegisterEvent("MERCHANT_CLOSED");
f:RegisterEvent("MERCHANT_UPDATE");

--Listens for addon load to setup slash commands
local function EventListener(self, event, ...)
	if (event == "ADDON_LOADED" and select(1, ...) == "ItemBuyer") then
		self:UnregisterEvent("ADDON_LOADED");
		
		StaticPopupDialogs["ItemBuyerConfirmation"] = {
			text = "Buy %d %s?", --numPurchases, item.link .. item.num
			button1 = "Confirm",
			button2 = "Cancel",
			OnAccept = function(self, data, data2)
				buyItem(data);
			end,
			timeout = 0,
			whileDead = false,
			hideOnEscape = true,
			preferredIndex = 3
		};
		
		SLASH_ITEMBUYER1 = "/itembuyer";
		SLASH_ITEMSELLER1 = "/itemseller";
		
		SlashCmdList["ITEMBUYER"] = itemBuyListener;
		SlashCmdList["ITEMSELLER"] = itemSellListener;
	elseif (event == "MERCHANT_SHOW") then
		if (not isAtMerchant) then
			msg("Merchant found!");
			isAtMerchant = true;
			merchantItems = getMerchantItems();
		end
	elseif (event == "MERCHANT_CLOSED") then
		if (isAtMerchant) then
			--remove any current confirmation since we arn't at the merchant anymore
			StaticPopup_Hide("ItemBuyerConfirmation");
			
			msg("Merchant lost!");
			isAtMerchant = false;
			merchantUpdated = false;
			merchantItems = {};
		end
	elseif (event == "MERCHANT_UPDATE") then
		if (not merchantUpdated) then
			merchantItems = getMerchantItems();
			
			if (merchantUpdated) then
				msg("Merchant Updated!");
			end
		end
	end
end

f:SetScript("OnEvent", EventListener);